from rest_framework.filters import SearchFilter
from rest_framework.viewsets import ModelViewSet, ReadOnlyModelViewSet


class AuditLogReadOnlyViewSet(ReadOnlyModelViewSet):

    audit_log_list_response = False

    def _get_lookup_kwargs(self):
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
        kwargs = getattr(self, 'kwargs', {})    # kwargs is set during dispatch(). Prevent possible attribute errors
        return {self.lookup_field: kwargs.get(lookup_url_kwarg, "")}

    def _get_filter_kwargs(self, request):
        search_terms = []
        for backend in list(self.filter_backends):
            if issubclass(backend, SearchFilter):
                search_terms += backend().get_search_terms(request)

        return {str(getattr(self, 'search_fields', [])): search_terms}

    def _get_model_name(self, queryset=None):
        if queryset is None:
            queryset = self.get_queryset()
        return queryset.model.__name__

    def retrieve(self, request, *args, **kwargs):
        response = super().retrieve(request, *args, **kwargs)

        if hasattr(request, 'audit_log'):
            model = self._get_model_name()
            request.audit_log.set_filter(object_name=model, kwargs=self._get_lookup_kwargs())
            request.audit_log.set_results(response.data)
            request.audit_log.info("Retrieve %s" % model)

        return response

    def list(self, request, *args, **kwargs):
        response = super().list(request, *args, **kwargs)

        if hasattr(request, 'audit_log'):
            model = self._get_model_name()
            filter_kwargs = self._get_filter_kwargs(request)
            request.audit_log.set_filter(object_name=model, kwargs=filter_kwargs)
            request.audit_log.info("List %s" % model)

            if self.audit_log_list_response:
                request.audit_log.set_results(response.data)

        return response


class AuditLogViewSet(AuditLogReadOnlyViewSet, ModelViewSet):

    def create(self, request, *args, **kwargs):
        response = super().create(request, *args, **kwargs)

        if hasattr(request, 'audit_log'):
            model = self._get_model_name()
            request.audit_log.set_results(response.data)
            request.audit_log.info("Created %s object" % model)

        return response

    def update(self, request, *args, **kwargs):
        partial = kwargs.get('partial', False)
        response = super().update(request, *args, **kwargs)

        if hasattr(request, 'audit_log'):
            model = self._get_model_name()
            message = "Partial update of %s" % model if partial else "Update of %s" % model
            request.audit_log.set_filter(object_name=model, kwargs=self._get_lookup_kwargs())
            request.audit_log.set_results(response.data)
            request.audit_log.info(message)

        return response

    def destroy(self, request, *args, **kwargs):
        response = super().destroy(request, *args, **kwargs)

        if hasattr(request, 'audit_log'):
            model = self._get_model_name()
            request.audit_log.set_filter(object_name=model, kwargs=self._get_lookup_kwargs())
            request.audit_log.set_results(response.data)
            request.audit_log.info("Destroy %s" % model)

        return response



import importlib
import logging

from django.http import HttpRequest


def get_client_ip(request: HttpRequest) -> str:
    try:
        x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
        if x_forwarded_for:
            ip = x_forwarded_for.split(',')[0]
        else:
            ip = request.META.get('REMOTE_ADDR')
        return ip
    except Exception:
        logger = logging.getLogger(__name__)
        logger.warning('Failed to get ip for audit log', exc_info=True)
        return 'failed to get ip'


def import_callable(path):
    module, method = path.rsplit('.', 1)
    return getattr(importlib.import_module(module), method)




from django.conf import settings

# Internal logger name used by the audit log. Leave None to use python-audit-log default ('audit_log')
LOGGER_NAME = getattr(settings, 'AUDIT_LOG_LOGGER_NAME', None)

# Log handler that determines what to do with the logs. Leave None to use python-audit-log default (write to stdout)
LOG_HANDLER_CALLABLE_PATH = getattr(settings, 'AUDIT_LOG_HANDLER_CALLABLE_PATH', None)

# Log formatter that determines log formatting. Leave None to use python-audit-log default (AuditLogFormatter)
LOG_FORMATTER_CALLABLE_PATH = getattr(settings, 'AUDIT_LOG_FORMATTER_CALLABLE_PATH', None)

# List of url names that will not be logged (e.g. health urls)
# Default: [] (Empty list)

# If a URL path matches a regular expression in this list, the request will not be redirected to HTTPS.
# The AuditLogMiddleware strips leading slashes from URL paths, so patterns shouldn’t include them, e.g.
# AUDIT_LOG_EXEMPT_URLS = [r'^foo/bar/$', …].
EXEMPT_URLS = getattr(settings, 'AUDIT_LOG_EXEMPT_URLS', [])

assert type(EXEMPT_URLS) is list, "EXEMPT_URLS must be a list"



import re

from django.http import HttpRequest, HttpResponse
from django.utils.deprecation import MiddlewareMixin

from django_audit_log import app_settings
from django_audit_log.logger import DjangoAuditLogger


class AuditLogMiddleware(MiddlewareMixin):

    def __init__(self, get_response=None):
        super().__init__(get_response)
        self.redirect_exempt = [re.compile(r) for r in app_settings.EXEMPT_URLS]

    def process_request(self, request: HttpRequest) -> None:
        if not hasattr(request, 'audit_log') and not self.exempt_request(request):
            audit_log = DjangoAuditLogger()
            audit_log.set_django_http_request(request)
            audit_log.set_user_from_request(request)
            request.audit_log = audit_log

    def process_response(self, request: HttpRequest, response: HttpResponse) -> HttpResponse:
        if hasattr(request, 'audit_log'):
            audit_log = request.audit_log
            audit_log.set_django_http_response(response)
            audit_log.send_log()
        return response

    def exempt_request(self, request):
        path = request.path.lstrip("/")
        return any(pattern.search(path) for pattern in self.redirect_exempt)



import logging

from django.http import HttpRequest, HttpResponse

from audit_log.logger import AuditLogger
from django_audit_log import app_settings
from django_audit_log.util import get_client_ip, import_callable


class DjangoAuditLogger(AuditLogger):

    def get_logger_name(self) -> str:
        logger_name = app_settings.LOGGER_NAME
        if not logger_name:
            return super().get_logger_name()
        return logger_name

    def get_log_handler(self) -> logging.Handler:
        log_handler_path = app_settings.LOG_HANDLER_CALLABLE_PATH
        if not log_handler_path:
            return super().get_log_handler()

        log_handler_callable = import_callable(log_handler_path)
        return log_handler_callable()

    def get_log_formatter(self) -> logging.Formatter:
        log_formatter_path = app_settings.LOG_FORMATTER_CALLABLE_PATH
        if not log_formatter_path:
            return super().get_log_formatter()

        log_formatter_callable = import_callable(log_formatter_path)
        return log_formatter_callable()

    def set_django_http_request(self, request: HttpRequest) -> 'DjangoAuditLogger':
        self.set_http_request(
            method=request.method,
            url=request.build_absolute_uri(),
            user_agent=request.META.get('HTTP_USER_AGENT', '?') if request.META else '?'
        )
        return self

    def set_django_http_response(self, response: HttpResponse) -> 'DjangoAuditLogger':
        headers = self._get_headers_from_response(response)
        self.set_http_response(
            status_code=getattr(response, 'status_code', ''),
            reason=getattr(response, 'reason_phrase', ''),
            headers=headers
        )
        return self

    def set_user_from_request(self, request: HttpRequest, realm='') -> 'DjangoAuditLogger':
        user = request.user if hasattr(request, 'user') else None
        roles = list(user.groups.values_list('name', flat=True)) if user else []
        self.set_user(
            authenticated=user.is_authenticated if user else False,
            provider=request.session.get('_auth_user_backend', '') if hasattr(request, 'session') else '',
            realm=realm,
            email=getattr(user, 'email', '') if user else '',
            roles=roles,
            ip=get_client_ip(request)
        )
        return self

    def _get_headers_from_response(self, response: HttpResponse) -> dict:
        return {header: value for header, value in response.items()}



import os
import sys

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

import django_audit_log


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


install_requirements = [
    'django',
    'datapunt-audit-log'
]

test_requirements = [
    'pytest',
    'pytest-cov'
]

extra_requirements = {
    'dev': test_requirements + ['twine', 'bump2version'],
}

setup(
    name='datapunt-django-audit-log',
    version=django_audit_log.__version__,
    license='Mozilla Public License 2.0',

    author='Datapunt Amsterdam',
    author_email='datapunt@amsterdam.nl',

    description='A simple Django app to enable audit logging',
    long_description=README,
    long_description_content_type="text/markdown",
    url='https://github.com/Amsterdam/auditlog',

    packages=find_packages(),
    install_requires=install_requirements,

    cmdclass={'test': PyTest},
    tests_require=test_requirements,

    extras_require=extra_requirements,

    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: System :: Logging'
    ],
)
